#ifndef LIGO_CONTROLLER_H
#define LIGO_CONTROLLER_H

#include "cds_types.h" //adcInfo_t, dacInfo_t, timing_diag_t
#include "fm10Gen_types.h" //FILT_MOD 
#include "daq_core_defs.h" //DAQ_DCU_SIZE

#include "drv/shmem.h"
#include "../shmem/shmem_all.h" //TESTPOINT_CFG, AWG_DATA


extern adcInfo_t     adcinfo;
extern dacInfo_t     dacinfo;
extern timing_diag_t timeinfo;

/// Maintains present cycle count within a one second period.
extern int          cycleNum;
extern unsigned int odcStateWord;
/// Value of readback from DAC FIFO size registers; used in diags for FIFO
/// overflow/underflow.
extern int          out_buf_size; // test checking DAC buffer size
extern unsigned int cycle_gps_time; // Time at which ADCs triggered
extern unsigned int cycle_gps_event_time; // Time at which ADCs triggered
extern unsigned int cycle_gps_ns;
extern unsigned int cycle_gps_event_ns;
extern unsigned int gps_receiver_locked; // Lock/unlock flag for GPS time card
/// GPS time in GPS seconds
extern unsigned int     timeSec;
extern unsigned int     timeSecDiag;
extern unsigned int     ipcErrBits;
extern int              cardCountErr;
extern struct rmIpcStr* daqPtr;
extern int dacOF[ MAX_DAC_MODULES ]; /// @param dacOF[]  DAC overrange counters

extern char daqArea[ DAQ_DCU_SIZE ]; // Space allocation for daqLib buffers
extern int  cpuId;

// Initial diag reset flag
extern int initialDiagReset;

// Cache flushing mumbo jumbo suggested by Thomas Gleixner, it is probably
// useless Did not see any effect
extern char fp[ 64 * 1024 ];

#define MX_OK 3


// The following bits define the EPICS STATE_WORD
#define FE_ERROR_TIMING 0x2 // bit 1
#define FE_ERROR_ADC 0x4 // bit 2
#define FE_ERROR_DAC 0x8 // bit 3
#define FE_ERROR_DAQ 0x10 // bit 4
#define FE_ERROR_IPC 0x20 // bit 5
#define FE_ERROR_AWG 0x40 // bit 6
#define FE_ERROR_DAC_ENABLE 0x80 // bit 7
#define FE_ERROR_EXC_SET 0x100 // bit 8
#define FE_ERROR_OVERFLOW 0x200 // bit 9
#define FE_ERROR_CFC 0x400 // bit 10, used by and also defined in skeleton.st

#define ODC_ADC_OVF 0x1
#define ODC_DAC_OVF 0x2
#define ODC_EXC_SET 0x4

#define CPURATE ( cpu_khz / 1000 )
#define ONE_PPS_THRESH 1000
#define SYNC_SRC_NONE 0
#define SYNC_SRC_DOLPHIN 1
#define SYNC_SRC_1PPS 2
#define SYNC_SRC_TDS 4
#define SYNC_SRC_MASTER 8
#define SYNC_SRC_TIMER 16
#define SYNC_SRC_LTC 32
#define SYNC_SRC_INTERNAL 64

#define TIME_ERR_IRIGB 0x10
#define TIME_ERR_1PPS 0x20
#define TIME_ERR_TDS 0x40

#define CPU_TIMER_CNT 10
#define CPU_TIME_CYCLE_START 0
#define CPU_TIME_CYCLE_END 1
#define CPU_TIME_USR_START 4
#define CPU_TIME_USR_END 5
#define CPU_TIME_RDY_ADC 8
#define CPU_TIME_ADC_WAIT 9

// fe_state defs
#define IO_CARD_MAP_ERROR -9
#define IO_CONFIG_ERROR -8
#define ADC_TO_ERROR -7
#define DAC_INIT_ERROR -6
#define BURT_RESTORE_ERROR -5
#define CHAN_HOP_ERROR -4
#define RUN_ON_TIMER -3
#define DAQ_INIT_ERROR -2
#define FILT_INIT_ERROR -1
#define LOADING 0
#define FIND_MODULES 1
#define WAIT_BURT 2
#define LOCKING_CORE 3
#define INIT_ADC_MODS 4
#define INIT_DAC_MODS 5
#define INIT_SYNC 6
#define NORMAL_RUN 7
#define IOC_SLOT_CHK 8
#define FE_CODE_INIT    1
#define FE_CODE_PROCESS    0

// Define standard values based on code rep rate
// **************************************
#ifdef SERVO512K
#define DAQ_CYCLE_CHANGE ( 4 * 8000 )
#define END_OF_DAQ_BLOCK 32767
#define DAQ_RATE ( DAQ_16K_SAMPLE_SIZE * 32 )
#define CYCLE_TIME_ALRM 15
#define CYCLE_TIME_ALRM_HI 25
#define CYCLE_TIME_ALRM_LO 10
#define DAC_PRELOAD_CNT 0
#define IOC_CLK_SLOW	0x10
#define IOC_CLK_FAST	0x13
#endif
#ifdef SERVO256K
#define DAQ_CYCLE_CHANGE ( 2 * 8000 )
#define END_OF_DAQ_BLOCK 16383
#define DAQ_RATE ( DAQ_16K_SAMPLE_SIZE * 16 )
#define CYCLE_TIME_ALRM 4
#define CYCLE_TIME_ALRM_HI 5
#define CYCLE_TIME_ALRM_LO 1
#define DAC_PRELOAD_CNT 0
#define IOC_CLK_SLOW	0x10
#define IOC_CLK_FAST	0x12
#endif
#ifdef SERVO128K
#define DAQ_CYCLE_CHANGE 8000
#define END_OF_DAQ_BLOCK 8191
#define DAQ_RATE ( DAQ_16K_SAMPLE_SIZE * 8 )
#define CYCLE_TIME_ALRM 7
#define CYCLE_TIME_ALRM_HI 12
#define CYCLE_TIME_ALRM_LO 1
#define DAC_PRELOAD_CNT 0
#define IOC_CLK_SLOW	0x11
#define IOC_CLK_FAST	0x11
#endif
#ifdef SERVO64K
#define DAQ_CYCLE_CHANGE ( 2 * 1540 )
#define END_OF_DAQ_BLOCK 4095
#define DAQ_RATE ( DAQ_16K_SAMPLE_SIZE * 4 )
#ifdef CONTROL_MODEL
#define CYCLE_TIME_ALRM 18
#define CYCLE_TIME_ALRM_HI 23
#define CYCLE_TIME_ALRM_LO 10
#else
#define CYCLE_TIME_ALRM         (18 * (UNDERSAMPLE / ADC_MEMCPY_RATE))
#define CYCLE_TIME_ALRM_HI      (23 * (UNDERSAMPLE / ADC_MEMCPY_RATE))
#define CYCLE_TIME_ALRM_LO      10
#endif
#define IOC_CLK_SLOW	0x10
#define IOC_CLK_FAST	0x10
#ifdef CONTROL_MODEL
#define DAC_PRELOAD_CNT 1
#else
#define DAC_PRELOAD_CNT 0
#endif
#endif
#ifdef SERVO32K
#define DAQ_CYCLE_CHANGE 1540
#define END_OF_DAQ_BLOCK 2047
#define DAQ_RATE ( DAQ_16K_SAMPLE_SIZE * 2 )
#define CYCLE_TIME_ALRM_HI 38
#define CYCLE_TIME_ALRM_LO 25
#define CYCLE_TIME_ALRM 31
#define IOC_CLK_SLOW	0x10
#define IOC_CLK_FAST	0x10
#ifdef CONTROL_MODEL
#define DAC_PRELOAD_CNT 2
#else
#define DAC_PRELOAD_CNT 1
#endif
#endif

#ifdef SERVO16K
#define DAQ_CYCLE_CHANGE 770
#define END_OF_DAQ_BLOCK 1023
#define DAQ_RATE DAQ_16K_SAMPLE_SIZE
#define CYCLE_TIME_ALRM_HI 70
#define CYCLE_TIME_ALRM_LO 50
#define CYCLE_TIME_ALRM 62
#define DAC_PRELOAD_CNT 4
#endif
#ifdef SERVO4K
#define DAQ_CYCLE_CHANGE 240
#define END_OF_DAQ_BLOCK 255
#define DAQ_RATE 2 * DAQ_2K_SAMPLE_SIZE
#define CYCLE_TIME_ALRM 487 / 2
#define CYCLE_TIME_ALRM_HI 500 / 2
#define CYCLE_TIME_ALRM_LO 460 / 2
#define DAC_PRELOAD_CNT 16
#endif
#ifdef SERVO2K
#define DAQ_CYCLE_CHANGE 120
#define END_OF_DAQ_BLOCK 127
#define DAQ_RATE DAQ_2K_SAMPLE_SIZE
#define CYCLE_TIME_ALRM 487
#define CYCLE_TIME_ALRM_HI 500
#define CYCLE_TIME_ALRM_LO 460
#define DAC_PRELOAD_CNT 8
#endif

// DIAGNOSTIC_RETURNS_FROM_FE
#define FE_NO_ERROR 0x0
#define FE_SYNC_ERR 0x1
#define FE_ADC_HOLD_ERR 0x2
#define FE_FB0_NOT_ONLINE 0x4
#define FE_PROC_TIME_ERR 0x8
#define FE_ADC_SYNC_ERR 0xf0
#define FE_FB_AVAIL 0x1
#define FE_FB_ONLINE 0x2
#define FE_MAX_FB_QUE 0x10
#define ADC_TIMEOUT_ERR 0x1

// DAC status bits to EPICS
#define DAC_FOUND_BIT 1
#define DAC_TIMING_BIT 2
#define DAC_OVERFLOW_BIT 4
#define DAC_FIFO_BIT 8
#define DAC_WD_BIT 16
#define DAC_FIFO_EMPTY 32
#define DAC_FIFO_HI_QTR 64
#define DAC_FIFO_FULL 128
#define DAC_AUTOCAL_BIT 256
#define DAC_WD_TRIP_SET 5

// microseconds per IOP cycle, *NOT* microseconds per fastes ADC.
#define USEC_PER_CYCLE (1000000 * UNDERSAMPLE / MODEL_RATE_CPS )

#define MAX_IRIGB_SKEW  (int)((USEC_PER_CYCLE) * 1.6)
#define MIN_IRIGB_SKEW (int)((USEC_PER_CYCLE) * 0.334)
#define DT_SAMPLE_OFFSET 6
#define DT_SAMPLE_CNT 12
#define MAX_DT_DIAG_VAL 6
#define MIN_DT_DIAG_VAL 4

// HOUSEKEEPING CYCLE DEFS
// 1Hz Jobs triggered by cycleNum count in controller code
#define HKP_READ_SYMCOM_IRIGB 0
#define HKP_READ_TSYNC_IRIBB 1
#define HKP_READ_DIO 10
#define HKP_DT_CALC 16
#define HKP_DAC_DT_SWITCH 17
#define HKP_TIMING_UPDATES 18
#define HKP_DIAG_UPDATES 19
#define HKP_DAC_EPICS_UPDATES 20
#define HKP_ADC_DAC_STAT_UPDATES 21
#define HKP_RFM_CHK_CYCLE 300 // ONLY IOP
#define HKP_DAC_WD_CLK 400 // ONLY IOP
#define HKP_DAC_WD_CHK 500 // ONLY IOP
#define HKP_DAC_FIFO_CHK 600 // ONLY IOP
#define HKP_DAC_FIFO_CHK 600 // ONLY IOP
#define HKP_IOC_SLOT_STATUS 1000 // ONLY IOP
#define HKP_IOC_SLOT_CONFIG 1100 // ONLY IOP
#define HKP_IOC_BP_CONFIG 1200 // ONLY IOP
#define HKP_IOC_BP_STATUS 1205 // ONLY IOP
#define HKP_IOC_LPTC_STATUS 1210 // ONLY IOP
// 16Hz Jobs triggered by subcycle count in controller code
#define HKP_FM_EPICS_UPDATE 30

#define NUM_SYSTEMS 1
#define INLINE inline
#define MMAP_SIZE ( 64 * 1024 * 1024 - 5000 )

#define STACK_SIZE 40000
#define TICK_PERIOD 100000
#define PERIOD_COUNT 1
#define MAX_UDELAY 19999


extern char*          build_date ;

extern volatile char* _epics_shm; ///< Ptr to EPICS shared memory area
extern shmem_handle g_epics_shm_handle;

extern volatile TESTPOINT_CFG  *_tp_shm;
extern shmem_handle g_tp_shm_handle;

extern volatile AWG_DATA *_awg_shm;
extern shmem_handle g_awg_shm_handle;

extern volatile char*          _ipc_shm; ///< Ptr to inter-process communication area
extern shmem_handle g_ipc_shm_handle;

extern volatile char*          _daq_shm; ///< Ptr to frame builder comm shared mem area
extern shmem_handle g_daq_shm_handle;

extern volatile char*          _shmipc_shm; ///< Ptr to IOP I/O data to/from User app shared mem area
extern shmem_handle g_shmipc_shm_handle;

extern volatile char*          _io_shm; ///< Ptr to user space I/O area
extern shmem_handle g_io_shm_handle;

extern int            daq_fd; ///< File descriptor to share memory file

extern long                      daqBuffer; // Address for daq dual buffers in daqLib.c
extern CDS_HARDWARE              cdsPciModules; // Structure of PCI hardware addresses
extern volatile IO_MEM_DATA*     ioMemData;
extern volatile IO_MEM_DATA_IOP* ioMemDataIop;
extern volatile int              vmeDone ; // Code kill command
extern volatile int              fe_status_return; // fe code status return to module_exit
extern volatile int              fe_status_return_subcode; // fe code status return to module_exit
extern volatile int              stop_working_threads;

extern int     iop_rfm_valid; //From mbuf kernel module
extern unsigned int cpu_khz; //From linux kernel tsc.c

// Following define string names of above modules
// Used to print info to syslog
#define CDS_CARD_TYPES  16
extern char _cdscardtypename[ CDS_CARD_TYPES ][ 40 ];


/// Testpoints which are not part of filter modules
extern double* testpoint[ GDS_MAX_NFM_TP ];
#ifndef NO_DAQ
extern DAQ_RANGE daq; // Range settings for daqLib.c
extern int       numFb;
extern int       fbStat[ 2 ]; // Status of DAQ backend computer
/// Excitation points which are not part of filter modules
extern double xExc[ GDS_MAX_NFM_EXC ]; // GDS EXC not associated with filter modules
#endif
/// 1/16 sec cycle counters for DAQS
extern int subcycle; // Internal cycle counter
/// DAQ cycle counter (0-15)
extern unsigned int daqCycle; // DAQS cycle counter

// Sharded memory discriptors
extern int wfd;
extern int ipc_fd;
extern volatile CDS_EPICS* pLocalEpics; // Local mem ptr to EPICS control data
extern volatile char*      pEpicsDaq; // Local mem ptr to EPICS daq data

// Filter module variables
/// Standard Filter Module Structure
extern FILT_MOD dsp[ NUM_SYSTEMS ]; // SFM structure.
/// Pointer to local memory SFM Structure
extern FILT_MOD* dspPtr[ NUM_SYSTEMS ]; // SFM structure pointer.
/// Pointer to SFM in shared memory.
extern FILT_MOD* pDsp[ NUM_SYSTEMS ]; // Ptr to SFM in shmem.
/// Pointer to filter coeffs local memory.
extern COEF dspCoeff[ NUM_SYSTEMS ]; // Local mem for SFM coeffs.
/// Pointer to filter coeffs shared memory.
extern VME_COEF* pCoeff[ NUM_SYSTEMS ]; // Ptr to SFM coeffs in shmem

// ADC Variables
/// Array of ADC values
#ifdef IOP_MODEL
extern double dWord[ MAX_ADC_MODULES ][ MAX_ADC_CHN_PER_MOD ][ 16 ]; // ADC read values
#else
extern double dWord[ MAX_ADC_MODULES ][ MAX_ADC_CHN_PER_MOD ]; // ADC read values
#endif
/// List of ADC channels used by this app. Used to determine if downsampling
/// required.
extern unsigned int dWordUsed[ MAX_ADC_MODULES ]
                             [ MAX_ADC_CHN_PER_MOD ]; // ADC chans used by app code

// DAC Variables
/// Enables writing of DAC values; Used with DACKILL parts..
extern int iopDacEnable; // Returned by feCode to allow writing values or zeros to DAC
                  // modules
extern int dacChanErr[ MAX_DAC_MODULES ];
#ifdef IOP_MODEL
extern int dacOutBufSize[ MAX_DAC_MODULES ];
#endif
/// Array of DAC output values.
extern double dacOut[ MAX_DAC_MODULES ][ MAX_DAC_CHN_PER_MOD ]; // DAC output values
/// DAC output values returned to EPICS
extern int dacOutEpics[ MAX_DAC_MODULES ]
                      [ MAX_DAC_CHN_PER_MOD ]; // DAC outputs reported back to EPICS
/// DAC channels used by an app.; determines up sampling required.
extern unsigned int dacOutUsed[ MAX_DAC_MODULES ]
                              [ MAX_DAC_CHN_PER_MOD ]; // DAC chans used by app code
/// Array of DAC overflow (overrange) counters.
extern int overflowDac[ MAX_DAC_MODULES ]
                      [ MAX_DAC_CHN_PER_MOD ]; // DAC overflow diagnostics
/// DAC outputs stored as floats, to be picked up as test points
extern double floatDacOut[ 160 ]; // DAC outputs stored as floats, to be picked up as
                           // test points

/// Counter for total ADC/DAC overflows
extern int overflowAcc; // Total ADC/DAC overflow counter

#ifndef IOP_MODEL
// Variables for Digital I/O board values
// DIO board I/O is handled in control (user) applications for timing reasons
// (longer I/O access times)
/// Read value from Acces I/O 24bit module
extern int dioInput[ MAX_DIO_MODULES ];
/// Write value to Acces I/O 24bit module
extern int dioOutput[ MAX_DIO_MODULES ];
/// Last value written to Acces I/O 24bit module
extern int dioOutputHold[ MAX_DIO_MODULES ];

extern int rioInputOutput[ MAX_DIO_MODULES ];
extern int rioOutput[ MAX_DIO_MODULES ];
extern int rioOutputHold[ MAX_DIO_MODULES ];

extern int rioInput1[ MAX_DIO_MODULES ];
extern int rioOutput1[ MAX_DIO_MODULES ];
extern int rioOutputHold1[ MAX_DIO_MODULES ];
extern int rioInputInput[ MAX_DIO_MODULES ];


// Contec 32 bit output modules
/// Read value from Contec 32bit I/O module
extern unsigned int CDO32Input[ MAX_DIO_MODULES ];
/// Write value to Contec 32bit I/O module
extern unsigned int CDO32Output[ MAX_DIO_MODULES ];


#endif

// Up/Down Sampling Filter Coeffs
// All systems not running at 64K require up/down sampling to communicate I/O
// data with IOP, which is running at 64K. Following defines the filter coeffs
// for these up/down filters.
#ifdef OVERSAMPLE
/* Recalculated filters in biquad form */

/* Oversamping base rate is 64K */
/* Coeffs for the 2x downsampling (32K system) filter */
extern double __attribute__( ( unused ) ) feCoeff2x[ 9 ];

/* RCG V3.0 Coeffs for the 4x downsampling (16K system) filter */
extern double __attribute__( ( unused ) ) feCoeff4x[ 9 ];

//
// New Brian Lantz 4k decimation filter
extern double __attribute__( ( unused ) ) feCoeff16x[ 9 ];


/* Coeffs for the 32x downsampling filter (2K system) per Brian Lantz May 5,
 * 2009 */
extern double __attribute__( ( unused ) ) feCoeff32x[ 9 ];

// History buffers for oversampling filters
extern double dHistory[ ( MAX_ADC_MODULES * 32 ) ][ MAX_HISTRY ];
extern double dDacHistory[ ( MAX_DAC_MODULES * 16 ) ][ MAX_HISTRY ];

#else

#define OVERSAMPLE_TIMES 1

#endif //OVERSAMPLE

#endif //LIGO_CONTROLLER_H
